/*
 *  code to build an expression "tree".
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sm-exprh.c,v 1.16 2013/10/23 22:13:41 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/io.h"
#include "sm/sm-conf.h"
#include "sm/sm-conf-expr.h"
#include "sm-expr.h"
#include "cond.h"

/* max. number of nodes in an expression */
#define EXPRN_MAX	256

sm_ret_T
m_expr_init(uint n, cexpr_P *pexpr)
{
	cexn_P exns;
	cexpr_P expr;

	SM_REQUIRE(n > 0);
	SM_REQUIRE(pexpr != NULL);

	expr = sm_malloc(sizeof(*expr));
	if (expr == NULL)
		return sm_err_temp(ENOMEM);
	exns = sm_malloc(n * sizeof(*exns));
	if (exns == NULL) {
		sm_free(expr);
		return sm_err_temp(ENOMEM);
	}
	sm_memzero(expr, sizeof(*expr));
	sm_memzero(exns, n * sizeof(*exns));
	expr->cex_total = n;
	expr->cex_next = 0;
	expr->cex_nodes = exns;
	*pexpr = expr;
	return SM_SUCCESS;
}

static sm_ret_T
m_expr_grow(cexpr_P expr, uint n)
{
	cexn_P exns;

	SM_REQUIRE(n > 0);
	SM_REQUIRE(expr != NULL);

	exns = sm_realloc(expr->cex_nodes, sizeof(*exns) * n);
	if (exns == NULL)
		return sm_err_temp(ENOMEM);
	expr->cex_total = n;
	expr->cex_nodes = exns;
	return SM_SUCCESS;
}

sm_ret_T
m_expr_free(cexpr_P expr)
{
	cexn_P exns;

	if (expr == NULL)
		return SM_SUCCESS;
	exns = expr->cex_nodes;
	if (exns != NULL) {
		free(exns);	/* need to free content of each node? */
		expr->cex_nodes = NULL;
		expr->cex_total = 0;
		expr->cex_next = 0;
	}
	/* free expr itself? */
	return SM_SUCCESS;
}

sm_ret_T
m_expr_get_node(cexpr_P expr, uint *pexi)
{
	SM_REQUIRE(expr != NULL);
	SM_REQUIRE(pexi != NULL);

	if (expr->cex_next >= expr->cex_total) {
		int n;
		sm_ret_T ret;

		if (expr->cex_total >= EXPRN_MAX)
			return sm_err_perm(SM_E_FULL);
		n = expr->cex_total;
		ret = m_expr_grow(expr, 2 * expr->cex_total);
		if (sm_is_err(ret))
			return ret;

		/*
		 *  Clear out all newly allocated structures.
		 *  This requires that the size is doubled above!
		 */
		sm_memzero(&CXN(expr, n), n * sizeof(cexn_T));
	}

	*pexi = expr->cex_next++;
	return SM_SUCCESS;
}

int
crt_exp(lpctx_P lpctx, int nargs, op_T op, uint const *args)
{
	int i;
	sm_ret_T ret;
	uint exni;

	SM_REQUIRE(lpctx != NULL);
	SM_REQUIRE(nargs > 0);
	SM_REQUIRE(nargs <= CEN_NARGS_MAX);

	for (i = nargs - 1; i >= 0; i--)
		if (args[i] >= lpctx->lpc_expr->cex_total)
			goto fail;

	/* Allocate a new expression.  */
	ret = m_expr_get_node(lpctx->lpc_expr, &exni);
	if (sm_is_err(ret))
		goto fail;
	CXN(lpctx->lpc_expr, exni).cen_nargs = nargs;
	CXN(lpctx->lpc_expr, exni).cen_vt = vt_op;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_op = op;
	for (i = nargs - 1; i >= 0; i--)
		CXN(lpctx->lpc_expr, exni).cen_args[i] = args[i];
	return exni;

  fail:
	return -1;
}

int
crt_id(lpctx_P lpctx, id_P id, val_type_T vt)
{
	sm_ret_T ret;
	uint exni;

	ret = m_expr_get_node(lpctx->lpc_expr, &exni);
	if (sm_is_err(ret))
		goto fail;
	CXN(lpctx->lpc_expr, exni).cen_vt = vt_id;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_id.id_str = id->id_str;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_id.id_len = id->id_len;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_id.id_idx = id->id_idx;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_id.id_vt = vt;
	return exni;

  fail:
	return -1;
}

int
crt_str(lpctx_P lpctx, id_P id)
{
	sm_ret_T ret;
	uint exni;

	ret = m_expr_get_node(lpctx->lpc_expr, &exni);
	if (sm_is_err(ret))
		goto fail;
	CXN(lpctx->lpc_expr, exni).cen_vt = vt_str;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_id.id_str = id->id_str;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_id.id_len = id->id_len;
	return exni;

  fail:
	return -1;
}

int
crt_int(lpctx_P lpctx, int v)
{
	sm_ret_T ret;
	uint exni;

	ret = m_expr_get_node(lpctx->lpc_expr, &exni);
	if (sm_is_err(ret))
		goto fail;
	CXN(lpctx->lpc_expr, exni).cen_vt = vt_int;
	CXN(lpctx->lpc_expr, exni).cen_val.cv_int = v;
	return exni;

  fail:
	return -1;
}

int
crt_exp_1(lpctx_P lpctx, op_T op, uint right)
{
	uint args[1];

	SM_REQUIRE(right < lpctx->lpc_expr->cex_total);
	args[0] = right;
	return crt_exp(lpctx, 1, op, args);
}

int
crt_exp_2(lpctx_P lpctx, op_T op, uint left, uint right)
{
	uint args[2];

	SM_REQUIRE(right < lpctx->lpc_expr->cex_total);
	SM_REQUIRE(left < lpctx->lpc_expr->cex_total);
	args[0] = left;
	args[1] = right;
	return crt_exp(lpctx, 2, op, args);
}

int
crt_exp_3(lpctx_P lpctx, op_T op, uint bexp, uint tbranch, uint fbranch)
{
	uint args[3];

	SM_REQUIRE(bexp < lpctx->lpc_expr->cex_total);
	SM_REQUIRE(tbranch < lpctx->lpc_expr->cex_total);
	SM_REQUIRE(fbranch < lpctx->lpc_expr->cex_total);
	args[0] = bexp;
	args[1] = tbranch;
	args[2] = fbranch;
	return crt_exp(lpctx, 3, op, args);
}
