/*
 *  Code based on tls_fprint.c from postfix which has the following
 * LICENSE
 *	This software is free. You can do with it whatever you want.
 *	The original author kindly requests that you acknowledge
 *	the use of his software.
 * AUTHOR(S)
 *	Wietse Venema
 *	IBM T.J. Watson Research
 *	P.O. Box 704
 *	Yorktown Heights, NY 10598, USA
 *
 *	Viktor Dukhovni
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: tlsfp.c,v 1.2 2014/01/11 02:13:09 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/dns.h"
#include "sm/tls.h"
#include "sm/tlsdane.h"
#include "sm/str.h"

#if MTA_USE_TLS && MTA_TLS_DANE

/*
**  TLS_DATA_MD -- compute md (binary fingerprint) for data
**
**	Parameters:
**		buf -- data to hash
**		mdalg -- algorithm to use
**
**	Returns:
**		usual error code
**
**	Side Effects:
**		overwrites buf with md
*/

sm_ret_T
tls_data_md(sm_str_P buf, const char *mdalg)
{
	EVP_MD_CTX *mdctx;
	const EVP_MD *md;
	uchar md_buf[EVP_MAX_MD_SIZE];
	unsigned int md_len;

	SM_REQUIRE(buf != NULL);
	SM_REQUIRE(mdalg != NULL);
	md = EVP_get_digestbyname(mdalg);
	if (NULL == md)
		return sm_error_perm(SM_EM_TLS, SM_E_UNEXPECTED);

	mdctx = EVP_MD_CTX_create();
	if (EVP_DigestInit_ex(mdctx, md, NULL) != 1)
		return sm_error_perm(SM_EM_TLS, SM_E_UNEXPECTED);
	if (EVP_DigestUpdate(mdctx, (void *)sm_str_data(buf), sm_str_getlen(buf)) != 1)
		return sm_error_perm(SM_EM_TLS, SM_E_UNEXPECTED);
	if (EVP_DigestFinal_ex(mdctx, md_buf, &md_len) != 1)
		return sm_error_perm(SM_EM_TLS, SM_E_UNEXPECTED);
	EVP_MD_CTX_destroy(mdctx);

	sm_str_clr(buf);
	return sm_str_scatn(buf, md_buf, md_len);
}

/*
**  TLS_PKEY_MD -- compute md (binary fingerprint) of public key of a cert
**
**	Parameters:
**		cert -- cert for which to compute public key md
**		mdalg -- algorithm to use
**		md -- "string" in which md (binary!) is stored (output)
**
**	Returns:
**		usual error code
*/

sm_ret_T
tls_pkey_md(X509 *cert, const char *mdalg, sm_str_P md)
{
	int len;
	sm_ret_T ret;
	uchar *buf;

	SM_REQUIRE(md != NULL);
	SM_REQUIRE(cert != NULL);
	SM_REQUIRE(mdalg != NULL);

	len = i2d_X509_PUBKEY(X509_get_X509_PUBKEY(cert), NULL);

	/* what's an acceptable upper limit? */
	if (len <= 0 || len >= 8192)
		return sm_error_perm(SM_EM_TLS, SM_E_OVFLW_NS);
	ret = sm_str_space(md, len);
	if (sm_is_err(ret)) return ret;
	buf = sm_str_data(md);
	i2d_X509_PUBKEY(X509_get_X509_PUBKEY(cert), &buf);
	if (buf - sm_str_data(md) != len)
		return sm_error_perm(SM_EM_TLS, SM_E_UNEXPECTED);
	SM_STR_SETLEN(md, len);
	return tls_data_md(md, mdalg);
}

#endif /* MTA_USE_TLS && MTA_TLS_DANE */
